#include "stdafx.h"

//----- Maintain array of open dialogues
HWND * g_aWindows = NULL;		// array of open dialogue handles 
int g_iWinAlloc=0;				// size of g_aWindows
int g_iWinOpen=0;				// number of g_aWindows in use 


//----- Global Instance Handle 
HINSTANCE g_hInst;								
HINSTANCE GetInstance( ) { 
	return g_hInst;
	}
void SetInstance( HINSTANCE hInst ) { 
	g_hInst=hInst;
	}


//-----
bool GetOpenFileName( char * pchFile, const char * pchFilter, const char * pchExt, HWND hOwner, char * pchShortName, CString * pInitialDir  ) { 
	OPENFILENAME ofn;
	memset( &ofn, 0, sizeof(OPENFILENAME) );
	ofn.lStructSize = sizeof(OPENFILENAME);
	ofn.hwndOwner = hOwner;
	ofn.lpstrFilter = pchFilter;
	ofn.lpstrDefExt=pchExt;
	ofn.lpstrInitialDir = pInitialDir->Buffer();
	
	ofn.nMaxFile = _MAX_PATH;
	ofn.lpstrFile = (LPSTR) pchFile;
	ofn.nMaxFileTitle = 0;
	ofn.lpstrFileTitle = NULL;
	bool result = (0!=GetOpenFileName(&ofn));

	char *path_sep = strrchr(pchFile, '\\');
	if ( path_sep )
		pInitialDir->Set(pchFile, path_sep - pchFile);

	return result;
	}

//-----
bool GetSaveFileName( char * pchFile, const char * pchFilter, const char * pchExt, HWND hOwner, char * pchShortName, CString * pInitialDir  ) { 
	OPENFILENAME ofn;
	memset( &ofn, 0, sizeof(OPENFILENAME) );
	ofn.lStructSize = sizeof(OPENFILENAME);
	ofn.hwndOwner = hOwner;
	ofn.lpstrFilter = pchFilter;
	ofn.lpstrDefExt=pchExt;
	ofn.lpstrInitialDir = pInitialDir->Buffer();
	
	ofn.nMaxFile = _MAX_PATH;
	ofn.lpstrFile = (LPSTR) pchFile;
	ofn.nMaxFileTitle = 0;
	ofn.lpstrFileTitle = NULL;
	//ofn.Flags = OFN_OVERWRITEPROMPT ;
	bool result = (0!=GetSaveFileName(&ofn));

	char *path_sep = strrchr(pchFile, '\\');
	if ( path_sep )
		pInitialDir->Set(pchFile, path_sep - pchFile);

	return result;
	}

//-----------------------------------------------------------------------------
//-- CDlgExtra 
CDlgExtra :: CDlgExtra( int iID, HWND hWndParent, int iX0, int iY0, int iX1, int iY1, bool bDraggable ) { 
	m_hWndParent = hWndParent;
	m_iID=iID;
	SetRect( &m_rct, iX0,iY0, iX1, iY1 );
	m_bDraggable=bDraggable;
	}
	
void CDlgExtra :: Draw( HDC hdc ){ 
	//default display ... nothing
	}
	
bool CDlgExtra :: HitTest( int iX, int iY ){ 
	return iX>=m_rct.left && iX<=m_rct.right && iY>=m_rct.top && iY<=m_rct.bottom;
	}
	
bool CDlgExtra :: DragDown( int iX, int iY ){ 
	if( ! m_bDraggable || ! HitTest(iX,iY) )
		return false;
	m_ptDragDown.x=iX;
	m_ptDragDown.y=iY;
	return true;
	}
	
void CDlgExtra :: DragMove( int iX, int iY ){ 
	m_ptDragMove.x=iX;
	m_ptDragMove.y=iY;
	}
	
void CDlgExtra :: DragUp( int iX, int iY ){ 
	m_ptDragMove.x=iX;
	m_ptDragMove.y=iY;
	}

void CDlgExtra :: Invalidate() {
	RECT r;
	SetRect( &r, m_rct.left, m_rct.top, m_rct.right+1, m_rct.bottom+1 );
	InvalidateRect( m_hWndParent, &r, false );
	}

//-----------------------------------------------------------------------------
CDlgVertText :: CDlgVertText( int iID, HWND hWndParent, int iX0, int iY0, int iX1, int iY1, 
							const char * pchText, int iSize, COLORREF crText, COLORREF crFill ) 
		: CDlgExtra( iID, hWndParent, iX0, iY0, iX1, iY1, false ), m_cText( pchText ) { 
	m_crText = crText;
	m_crFill = crFill;
	m_iSize=iSize;
	}
	
void CDlgVertText :: Draw( HDC hdc ) { 
	HFONT fnt = CreateFont( -1 * m_iSize, 0, 900, 900, FW_REGULAR, 0, 0, 0, 
						DEFAULT_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, 
						PROOF_QUALITY /*DEFAULT_QUALITY*/, 
						FF_SWISS /*FF_ROMAN*/ | VARIABLE_PITCH, NULL );
	HFONT fntWas = (HFONT) SelectObject( hdc, fnt );
	COLORREF crTextColorWas = SetTextColor( hdc, m_crText );
	COLORREF crBkWas = SetBkColor( hdc, m_crFill );
	UINT uiAlignWas = SetTextAlign( hdc, TA_CENTER );

	ExtTextOut( hdc, m_rct.left, (m_rct.top+m_rct.bottom)/2, ETO_CLIPPED, &m_rct, m_cText.Buffer(), m_cText.Length(), NULL );

	if( uiAlignWas != GDI_ERROR ) 	
		SetTextAlign( hdc, uiAlignWas);

	SetBkColor( hdc, crBkWas );		
	SetTextColor( hdc, crTextColorWas );
	SelectObject( hdc, fntWas );
	DeleteObject( fnt );
	} 

//-----------------------------------------------------------------------------
// Mini vertical scroll bar 
const int MINISCR_TIPMIN = 5;				// for DlgMiniVScroll : distance from center pt. of tip to edges

CDlgMiniScroll :: CDlgMiniScroll( int iID, HWND hWndParent, 
				int iX0, int iY0, int iX1, int iY1, 
				double dMin, double dMax, double dLow, double dHigh, 
				bool bVert,
				COLORREF crFill, COLORREF crTips, COLORREF crBody, COLORREF crFrame ) 
		: CDlgExtra( iID, hWndParent, iX0, iY0, max(iX1,iX0+bVert?0:MINISCR_TIPMIN*4), max(iY0+bVert?MINISCR_TIPMIN*4:0,iY1), true) { 
	m_bVert=bVert;
	m_dMin=dMin;
	m_dMax=dMax;
	if( m_dMax<=m_dMin )
		m_dMax=m_dMin+0.0001;
	SetLowVal( min(dMax,max(dMin,dLow)) );
	SetHighVal( min(dMax,max(m_dLow,dHigh)) );
	m_iDragItem=0;
	m_crFill=crFill;
	m_crTips=crTips;
	m_crBody=crBody;
	m_crFrame=crFrame;
	}	

void CDlgMiniScroll :: Draw( HDC hdc ) { 
	// Note : careful with the rect bounds here - right and bottom edge are not drawn by api func so add 1 extra
	RECT rct;
	
	HPEN penFrame = CreatePen( 0, 0, m_crFrame );
	HBRUSH brFill = CreateSolidBrush( m_crFill );
	HBRUSH brBody = CreateSolidBrush( m_crBody );
	HBRUSH brTips = CreateSolidBrush( m_crTips );
	HBRUSH brWas = (HBRUSH) SelectObject( hdc, brFill );
	HPEN penWas = (HPEN) SelectObject( hdc, penFrame );
	
	// frame
	Rectangle( hdc, m_rct.left, m_rct.top, m_rct.right+1, m_rct.bottom+1 );
	
	// body
	if( m_bVert ) 
		SetRect( &rct, m_rct.left+1, m_iHigh, m_rct.right, m_iLow  );
	else 
		SetRect( &rct, m_iLow, m_rct.top+1, m_iHigh, m_rct.bottom );
	FillRect( hdc, &rct, brBody );
	
	// high value tip
	if( m_bVert ) 
		SetRect( &rct, m_rct.left+1, m_iHigh-MINISCR_TIPMIN, m_rct.right, m_iHigh+MINISCR_TIPMIN );
	else
		SetRect( &rct, m_iHigh-MINISCR_TIPMIN, m_rct.top+1, m_iHigh+MINISCR_TIPMIN, m_rct.bottom  );
	FillRect( hdc, &rct, brTips );
	
	// low value tip
	if( m_bVert ) 
		SetRect( &rct, m_rct.left+1, m_iLow-MINISCR_TIPMIN, m_rct.right, m_iLow+MINISCR_TIPMIN );
	else
		SetRect( &rct, m_iLow-MINISCR_TIPMIN, m_rct.top+1, m_iLow+MINISCR_TIPMIN, m_rct.bottom  );
	FillRect( hdc, &rct, brTips );
	
	//-- if tips overlap, draw midline
	if( m_bVert ) {
		if( m_iLow-MINISCR_TIPMIN <= m_iHigh+MINISCR_TIPMIN ) {
			MoveToEx( hdc, m_rct.left+1, (m_iHigh+m_iLow)/2, NULL );
			LineTo( hdc, m_rct.right-1, (m_iHigh+m_iLow)/2 );
			}
		}
	else { 
		if( m_iLow+MINISCR_TIPMIN >= m_iHigh-MINISCR_TIPMIN ) {
			MoveToEx( hdc, (m_iHigh+m_iLow)/2, m_rct.top+1,  NULL );
			LineTo( hdc, (m_iHigh+m_iLow)/2, m_rct.bottom-1 );
			}
		}
	
	SelectObject( hdc, brWas );
	SelectObject( hdc, penWas );
	DeleteObject( penFrame );
	DeleteObject( brFill );
	DeleteObject( brBody );
	DeleteObject( brTips );
	} 
	
double CDlgMiniScroll :: GetLowVal(){
	return m_dLow;
	}

double CDlgMiniScroll :: SetLowVal(double dVal, bool bPixelSet){
	m_dLow = min( max( dVal, m_dMin ), m_dMax );
	if( m_dHigh < m_dLow ) 
		SetHighVal( m_dLow, bPixelSet );
	if( bPixelSet )
		m_iLow = ValueToPixel( m_dLow );
	Invalidate();
	return m_dLow;
	}

double CDlgMiniScroll :: GetHighVal(){
	return m_dHigh;
	}

double CDlgMiniScroll :: SetHighVal(double dVal, bool bPixelSet){
	m_dHigh = min( max( dVal, m_dMin ), m_dMax );
	if( m_dLow > m_dHigh ) 
		m_dLow = m_dHigh;
	if( bPixelSet ) 
		m_iHigh = ValueToPixel( m_dHigh );
	Invalidate();
	return m_dHigh;
	}

int CDlgMiniScroll :: ValueToPixel( double dVal ) {
	if( m_bVert ) 
		return m_rct.bottom-1-MINISCR_TIPMIN - int(0.5+(dVal-m_dMin)/(m_dMax-m_dMin)*(m_rct.bottom-m_rct.top-1-2*MINISCR_TIPMIN));
	return m_rct.left+1+MINISCR_TIPMIN+int(0.5+(dVal-m_dMin)/(m_dMax-m_dMin)*(m_rct.right-m_rct.left-1-2*MINISCR_TIPMIN));
	}

double CDlgMiniScroll :: PixelToValue( int iVal ) {
	if( m_bVert ) 
		return m_dMin + (m_dMax-m_dMin) * (m_rct.bottom-MINISCR_TIPMIN-iVal) / (m_rct.bottom-m_rct.top-1-2*MINISCR_TIPMIN);
	return m_dMin + (m_dMax-m_dMin) * (iVal-m_rct.left-MINISCR_TIPMIN) / (m_rct.right-m_rct.left-1-2*MINISCR_TIPMIN);
	}

bool CDlgMiniScroll :: DragDown( int iX, int iY ){
	// use parent method to set m_ptDragDown
	if( ! CDlgExtra::DragDown( iX, iY ) )
		return false;
	
	// check if click on something useful
	// check 1/2 of top then btm then top then body ( to prevent unmoveable tips when overlapped at max )
	int iPos = m_bVert ? iY:iX;
	if( m_bVert																// on top half of high tip 
			? ( iPos>=m_iHigh-MINISCR_TIPMIN && iPos<=m_iHigh )		
			: ( iPos>=m_iHigh && iPos<=m_iHigh+MINISCR_TIPMIN) )			// on top half of high tip ( to prevent unmoveable tips when overlapped at max )
		m_iDragItem=1;
	else if( iPos>=m_iLow-MINISCR_TIPMIN && iPos<=m_iLow+MINISCR_TIPMIN )	// on low tip
		m_iDragItem=2;
	else if( iPos>=m_iHigh-MINISCR_TIPMIN && iPos<=m_iHigh+MINISCR_TIPMIN )	// on high tip
		m_iDragItem=1;
	else if( m_bVert														// on body // note : high pixel = low value
			? ( iPos>=m_iHigh && iPos<=m_iLow )	
			: ( iPos>=m_iLow && iPos<=m_iHigh ) )
		m_iDragItem = 3;
	else 
		return false;

	// limit range of drag op 
	if( m_iDragItem==1 && m_bVert ) {
		m_iDragMin = m_rct.top+MINISCR_TIPMIN+1;
		m_iDragMax = m_ptDragDown.y + ( m_iLow - m_iHigh - 1);
		}
	else if( m_iDragItem==1 && ! m_bVert ) { 
		m_iDragMin = m_ptDragDown.x - ( m_iHigh - m_iLow - 1 );
		m_iDragMax = m_rct.right-MINISCR_TIPMIN;
		}
	else if( m_iDragItem==2 && m_bVert ) { 
		m_iDragMin = m_ptDragDown.y - ( m_iLow - m_iHigh + 1);
		m_iDragMax = m_rct.bottom-MINISCR_TIPMIN;
		}
	else if( m_iDragItem==2 && ! m_bVert ) { 
		m_iDragMin = m_rct.left+1+MINISCR_TIPMIN;
		m_iDragMax = m_ptDragDown.x + ( m_iHigh - m_iLow - 1 ) ;
		}
	else if( m_iDragItem==3 && m_bVert ) { 
		m_iDragMin = m_ptDragDown.y + (m_rct.top+1+MINISCR_TIPMIN) - m_iHigh ;
		m_iDragMax = m_ptDragDown.y + (m_rct.bottom-MINISCR_TIPMIN) - m_iLow;
		}
	else if( m_iDragItem==3 && ! m_bVert ) { 
		m_iDragMin = m_ptDragDown.x + (m_rct.left+1+MINISCR_TIPMIN) - m_iLow ;
		m_iDragMax = m_ptDragDown.x + (m_rct.right-MINISCR_TIPMIN) - m_iHigh;
		}

	return true;
	}
	
void CDlgMiniScroll :: DragMove( int iX, int iY ){
	if( m_iDragItem == 0 )
		return;
		
	// use parent method to set m_ptDragMove
	CDlgExtra::DragMove( iX, iY );

	// limit extent of drag 
	int iLim = min( max( m_bVert ? m_ptDragMove.y : m_ptDragMove.x, m_iDragMin ), m_iDragMax );
	double dVal = PixelToValue( iLim );
	
	switch( m_iDragItem ) { 
		case 1 :
			m_iHigh = iLim;
			SetHighVal( max( m_dLow, dVal ), false );
			break;

		case 2 :
			m_iLow = iLim;
			SetLowVal( min( m_dHigh, dVal ), false );
			break;

		case 3 :
			m_iLow = m_bVert 
				? m_rct.bottom-MINISCR_TIPMIN - (m_iDragMax-iLim)
				: m_rct.left+1+MINISCR_TIPMIN + (iLim-m_iDragMin) ;
			m_iHigh = m_bVert 
				? m_rct.top+MINISCR_TIPMIN+1 + (iLim-m_iDragMin)
				: m_rct.right-MINISCR_TIPMIN - (m_iDragMax-iLim) ;
			SetLowVal( PixelToValue( m_iLow ), false );
			SetHighVal( PixelToValue( m_iHigh ), false );
			break;
			
		default : 
			return;
		}
	}
	
void CDlgMiniScroll :: DragUp( int iX, int iY ){
	// use parent method to set m_ptDragMove (not really necc.)
	CDlgExtra::DragUp( iX, iY );
	m_iDragItem = 0;
	}
	
//-----------------------------------------------------------------------------
//-- Dialog Control Base Class
CDlgControl :: CDlgControl( HWND h ) { 
	#ifdef _DEBUG
	if( h==NULL ) 
		handle=NULL;
	else
	#endif
		SetHandle(h);

	m_hTip=NULL;
	} 

void CDlgControl :: SetHandle( HWND h, char * pchTip ) { 
	handle=h; 
	#ifdef _DEBUG
	if( h == NULL ) 
		MessageBox( NULL, "Not a valid handle", "SetHandle", MB_OK );
	#endif
	
	if( h!=NULL && pchTip!=NULL )
		CreateTip( pchTip );
	}

HWND CDlgControl :: GetHandle() { 
	return handle; 
	} 

void CDlgControl :: SetEnabled( bool bEnable ){
	EnableWindow( handle, bEnable );
	}

void CDlgControl :: SetVisible( bool bVisible ){
	ShowWindow( handle, bVisible );
	}

void CDlgControl :: SetText( const char * pch ){
	SetWindowText( handle, pch );
	}

bool CDlgControl :: GetText( char * pch, int iMaxLen ){
	GetWindowText( handle, pch, iMaxLen );	// Note : return 0 can be error or empty text.
	return true;
	}

bool CDlgControl :: GetText( CString * pC ){
	int iLen = GetWindowTextLength( handle );
	pC->allocate( iLen+1 );
	GetWindowText( handle, pC->Buffer(), iLen+1 );
	pC->Buffer()[iLen]=0;
	pC->setlength();
	return true;
	}
	
int CDlgControl :: GetTextInt( ){
	int iLen = 1 + SendMessage( handle, WM_GETTEXTLENGTH, 0, 0 );
	char * buf = new char[iLen];
	SendMessage( handle, WM_GETTEXT, iLen, (LPARAM)buf );
	int iVal = atoi(buf);
	delete[]buf;
	return iVal;
	}
	
double CDlgControl :: GetTextDouble( ){
	int iLen = 1 + SendMessage( handle, WM_GETTEXTLENGTH, 0, 0 );
	char * buf = new char[iLen];
	SendMessage( handle, WM_GETTEXT, iLen, (LPARAM)buf );
	double dVal = atof(buf);
	delete[]buf;
	return dVal;
	}

void CDlgControl :: GetControlOffsetBounds( RECT * prct ) { 
	GetWindowRect( handle, prct );

	HWND hParent = GetParent( GetParent( handle ) );
	ScreenToClient( hParent, (POINT *)(prct) );
	ScreenToClient( hParent, (POINT *)(&prct->right) );

	InflateRect(prct, -1, -1);
	}

void CDlgControl :: GetControlBounds( RECT * prct ) { 
	GetWindowRect( handle, prct );

	HWND hParent = GetParent( handle );
	ScreenToClient( hParent, (POINT *)(prct) );
	ScreenToClient( hParent, (POINT *)(&prct->right) );

	InflateRect(prct, -1, -1);
	}

void CDlgControl :: MoveWindow( int iXOffset, int iYOffset ) { 
	RECT rct;
	GetWindowRect( handle, &rct );

	HWND hParent = GetParent( handle );
	ScreenToClient( hParent, (POINT *)(&rct) );
	ScreenToClient( hParent, (POINT *)(&(rct.right)) );

	int iX = rct.left+iXOffset;
	int iY = rct.top+iYOffset;

	SetWindowPos( handle, 0, iX, iY, 0, 0, SWP_NOSIZE | SWP_NOZORDER );
	}

void CDlgControl :: SizeWindow( int iWidth, int iHeight ) { 
	SetWindowPos( handle, 0, 0, 0, iWidth, iHeight, SWP_NOMOVE | SWP_NOZORDER );
	}

void CDlgControl :: SetTextInt( int iVal, char * pchFormat ) { 
	char buf[64];
	sprintf( buf, pchFormat, iVal );
	SetText( buf );
	}		
	
void CDlgControl :: SetTextDouble( double dVal, const char * pchFormat ) { 
	char buf[64];
	sprintf( buf, pchFormat==NULL ? "%f" : pchFormat, dVal );
	if( pchFormat==NULL  ) 
		ZTrim( buf );
	SetText( buf );
	}		

void CDlgControl :: SetButtonImage( HINSTANCE hInst, int iID, int iDX, int iDY ) { 
	HANDLE hi = LoadImage( GetInstance(), MAKEINTRESOURCE(iID), IMAGE_ICON, iDX, iDY, 0 );
	SendMessage( handle, BM_SETIMAGE, IMAGE_ICON, (LPARAM) hi );
	}

void CDlgControl :: SetButtonIcon( HINSTANCE hInst, int iID ) { 
	HICON hi = LoadIcon( hInst, MAKEINTRESOURCE(iID) );
	SendMessage( handle, BM_SETIMAGE, IMAGE_ICON, (LPARAM) hi );
	}

void CDlgControl :: SetButtonBitmap( HBITMAP hb ) { 
	SendMessage( handle, BM_SETIMAGE, IMAGE_BITMAP, (LPARAM) hb );
	}

void CDlgControl :: CreateTip( char * pchMsg, int iWidth ) { 
    // CREATE A TOOLTIP WINDOW 
    bool bNew = ( m_hTip == NULL );
    if( bNew ) { 
		m_hTip = CreateWindowEx( WS_EX_TOPMOST, TOOLTIPS_CLASS, NULL,
								WS_POPUP | TTS_NOPREFIX | TTS_ALWAYSTIP,		
								CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT,
								handle, NULL, GetInstance(), NULL );
		SetWindowPos( m_hTip, HWND_TOPMOST, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE | SWP_NOACTIVATE);
		}
	
	RECT rct;
	GetControlBounds( &rct );
	

    // add tip region
    TOOLINFO ti;
    ti.cbSize = sizeof(TOOLINFO);
    ti.uFlags = TTF_SUBCLASS;
    ti.hwnd = handle;
    ti.hinst = GetInstance();
    ti.uId = 0;
    ti.lpszText = pchMsg;       
    SetRect( &ti.rect, 0, 0, rct.right-rct.left, rct.bottom-rct.top );
	
    SendMessage( m_hTip, bNew ? TTM_ADDTOOL : TTM_UPDATETIPTEXT, 0, (LPARAM) (LPTOOLINFO) &ti );
    
    //-- if a width is specified, set it, which causes multiple lines to be available
    if( iWidth > 0 ) 
		SendMessage( m_hTip, TTM_SETMAXTIPWIDTH, 0, iWidth);
    
    }

//-----------------------------------------------------------------------------	
//-- Dialog edit control 	
void CDlgEdit :: SetTextLimit( int iLen ) { 
	SendMessage( handle, EM_SETLIMITTEXT, iLen, 0 );
	}
	
//-----------------------------------------------------------------------------	
//-- Dialog	checkbox control
void CDlgCheck :: SetCheck( bool bCheck ){
	SendMessage( handle, BM_SETCHECK, bCheck?BST_CHECKED:BST_UNCHECKED, 0);
	}

bool CDlgCheck :: GetCheck( ) {
	return BST_CHECKED == SendMessage( handle, BM_GETCHECK, 0, 0 );
	}

//-----------------------------------------------------------------------------	
void CDlgTab :: SetTab( int iTab ){
	SendMessage( handle, TCM_SETCURSEL, 0, 0 );
	}
int CDlgTab :: GetTab( ){
	return SendMessage( handle, TCM_GETCURSEL, 0, 0 );
	}
int CDlgTab :: GetTabCount( ){
	return SendMessage( handle, TCM_GETITEMCOUNT, 0, 0 );
	}
void CDlgTab :: DeleteTabs( ){
	SendMessage( handle, TCM_DELETEALLITEMS, 0, 0 );
	}
void CDlgTab :: DeleteTab( int iTab ){
	SendMessage( handle, TCM_DELETEITEM, iTab, 0 );
	}
bool CDlgTab :: AddTab( const char * pchName ){
	TCITEM tc;
	memset( &tc, 0, sizeof(tc) );
	tc.mask = TCIF_TEXT;		// also can set TCIF_IMAGE and TCIF_PARAM (cargo)
	tc.pszText = (LPSTR)pchName;
	return (-1!=SendMessage( handle, TCM_INSERTITEM, GetTabCount(), (LPARAM)(&tc) ) );	
	}

//-----------------------------------------------------------------------------	
//-- Dialog Scrollbar control
void CDlgScroll :: SetPos( int iPos ) { 
	SetScrollPos( handle, SB_CTL, iPos, TRUE );
	}

int CDlgScroll :: GetPos( ) { 
	return GetScrollPos( handle, SB_CTL );
	}

void CDlgScroll :: SetRange( int iTop, int iBottom, int iStepSize, int iPageSize ) { 
	SetScrollRange( handle, SB_CTL, m_iTop=iTop, m_iBottom=iBottom, true );
	m_iStepSize=iStepSize;
	m_iPageSize=iPageSize;
	}

int CDlgScroll :: ScrollProc( WPARAM wp ){ 
	switch( LOWORD( wp ) ) { 
		case SB_TOP :
			SetScrollPos( GetHandle(), SB_CTL, m_iTop, true );
			break;
		case SB_PAGEUP : 
			SetScrollPos( GetHandle(), SB_CTL, GetScrollPos( GetHandle(), SB_CTL )-m_iPageSize, true );
			break;
		case SB_LINEUP : 
			SetScrollPos( GetHandle(), SB_CTL, GetScrollPos( GetHandle(), SB_CTL )-m_iStepSize, true );
			break;
		case SB_LINEDOWN : 
			SetScrollPos( GetHandle(), SB_CTL, GetScrollPos( GetHandle(), SB_CTL )+m_iStepSize, true );
			break;
		case SB_PAGEDOWN : 
			SetScrollPos( GetHandle(), SB_CTL, GetScrollPos( GetHandle(), SB_CTL )+m_iPageSize, true );
			break;
		case SB_BOTTOM :
			SetScrollPos( GetHandle(), SB_CTL, m_iBottom, true );
			break;
		case SB_THUMBPOSITION :
			SetScrollPos( GetHandle(), SB_CTL, HIWORD(wp), true );
			break;
		default : 
			return -1;
		}
	return GetScrollPos( GetHandle(), SB_CTL );
	}

//-----------------------------------------------------------------------------	
void CDlgList :: ClearContents() {
	SendMessage( handle, LB_RESETCONTENT, 0, 0 );
	}
	
void CDlgList :: AddString( const char * pch ) {
	SendMessage( handle, LB_ADDSTRING, 0, (LPARAM) (LPCTSTR) pch );
	}
	
int CDlgList :: GetCurSel() {
	return (int) SendMessage( handle, LB_GETCURSEL, 0, 0 );		// LB_ERR (-1) if none
	}
	
int CDlgList :: GetCount() {
	return (int) SendMessage( handle, LB_GETCOUNT, 0, 0 );		
	}				
	
int CDlgList :: SetCurSel( int iSel ) {
	return (int) SendMessage( handle, LB_SETCURSEL, iSel, 0 );
	}	
	
bool CDlgList :: ReplaceString( int iSel, const char * pch ) {
	if( LB_ERR == SendMessage( handle, LB_DELETESTRING, iSel, 0 ) ) 
		return false;
	return LB_ERR != SendMessage( handle, LB_INSERTSTRING, iSel, (LPARAM)pch );			
	}


//-----------------------------------------------------------------------------	
//-- Dialog combo (droplist) control
void CDlgCombo :: ClearContents() { 
	SendMessage( handle, CB_RESETCONTENT, 0, 0 );
	}

void CDlgCombo :: AddString(const char * pch) { 
	SendMessage( handle, CB_ADDSTRING, 0, (LPARAM) (LPCTSTR) pch );
	}

void CDlgCombo :: InsertString(const char * pch, int iPos) { 
	SendMessage( handle, CB_INSERTSTRING, iPos, (LPARAM) (LPCTSTR) pch );
	}

void CDlgCombo :: DeleteString( int iSel ) { 
	SendMessage( handle, CB_DELETESTRING, iSel, 0 );
	}

int CDlgCombo :: FindString( const char * pch ) { 
	return (int) SendMessage( handle, CB_FINDSTRING, -1, (LPARAM)(LPCTSTR)pch );
	}

int CDlgCombo :: FindStringDouble( double dVal, const char * pchFormat ) { 
	char buf[64];
	sprintf( buf, pchFormat, dVal );
	return (int) SendMessage( handle, CB_FINDSTRING, -1, (LPARAM)(LPCTSTR)buf );
	}

int CDlgCombo :: FindStringInt( int iVal, const char * pchFormat ) { 
	char buf[64];
	sprintf( buf, pchFormat, iVal );
	return (int) SendMessage( handle, CB_FINDSTRING, -1, (LPARAM)(LPCTSTR)buf );
	}

int CDlgCombo :: GetCurSel() {
	return (int) SendMessage( handle, CB_GETCURSEL, 0, 0 );		// CB_ERR (-1) if none
	}

int CDlgCombo :: GetCount( ) {
	return (int) SendMessage( handle, CB_GETCOUNT, 0, 0 );		// CB_ERR (-1) if none
	}

bool CDlgCombo :: GetText( char * pch, int iMaxLen ) { 
	pch[0]=0;
	int iSel = (int) SendMessage( handle, CB_GETCURSEL, 0, 0 );
	if( iSel == CB_ERR )
		return false;
	int iLen = (int) SendMessage( handle, CB_GETLBTEXTLEN, iSel, 0 );	// does not include end 0
	char * achBuf = new char[iLen+1];
	if( achBuf==NULL ) 
		return false;
	SendMessage( handle, CB_GETLBTEXT, iSel, (LPARAM) (LPCTSTR) achBuf );
	if( iLen >= iMaxLen ) 
		achBuf[iMaxLen-1]=0;
	strcpy( pch, achBuf );
	delete [] achBuf;
	return true;
	}

int CDlgCombo :: SetCurSel( int iSel ) {
	return (int) SendMessage( handle, CB_SETCURSEL, iSel, 0 );		// CB_ERR (-1) if none
	}

int CDlgCombo :: SetCurSel( const char * pch, bool bAdd ) {
	// surprise!  CB_SELECTSTRING selects the first item that BEGINS WITH pch
	// int iPos = (int) SendMessage( handle, CB_SELECTSTRING, -1, (LPARAM) (LPCTSTR) pch );
	int iPos = (int) SendMessage( handle, CB_FINDSTRINGEXACT, -1, (LPARAM) (LPCTSTR) pch );
	if ( iPos >= 0 )
		iPos = SendMessage( handle, CB_SETCURSEL, iPos, (LPARAM)NULL );
	if( iPos == CB_ERR && bAdd ) { 
		SendMessage( handle, CB_ADDSTRING, 0, (LPARAM) (LPCTSTR) pch );
		iPos = (int) SendMessage( handle, CB_SELECTSTRING, -1, (LPARAM) (LPCTSTR) pch );
		}
	return iPos;
	}

bool CDlgCombo :: GetText( CString * pc ) { 
	//-- get cur sel 
	int iSel = GetCurSel();
	if( iSel == CB_ERR ) {
		CDlgControl :: GetText( pc );		
		return true;
		}
	return GetText( iSel, pc  );
	}

bool CDlgCombo :: GetText( int iItem, CString * pc ) { 
	//-- get selection length
	int iLen = (int) SendMessage( handle, CB_GETLBTEXTLEN, iItem, 0 );	// does not include end 0
	if( iLen==CB_ERR ) 
		return false;
	//-- allocate a buffer for string 
	char * achBuf = new char[iLen+1];
	if( achBuf==NULL ) 
		return false;
	//-- get text to buffer
	SendMessage( handle, CB_GETLBTEXT, iItem, (LPARAM) (LPCTSTR) achBuf );
	//-- convert to CString
	pc->Set( achBuf );
	//-- clear buffer
	delete [] achBuf;
	return true;
	}

	 	
//-----------------------------------------------------------------------------	
//-- Dialog 
CDialog :: CDialog() { 
	m_pCapture=NULL;
	m_hTip=NULL;
	m_bSizing=false;
	m_bDocumentDialog=false;
	}

void CDialog :: DrawExtras( HDC hdc ) {
	for( int iEx=0; iEx<m_aExtra.Count(); ++iEx )
		((CDlgExtra *)m_aExtra[iEx])->Draw(hdc);
	}

bool CDialog :: HitTest( int iX, int iY ){
	for( int iEx=0; iEx<m_aExtra.Count(); ++iEx )
		if( ((CDlgExtra *)m_aExtra[iEx])->HitTest( iX, iY ) )
			return true;
	return false;
	}

bool CDialog :: DragDown( int iX, int iY ){
	for( int iEx=0; iEx<m_aExtra.Count(); ++iEx )
		if( ((CDlgExtra *)m_aExtra[iEx])->DragDown(iX, iY ) ) { 
			m_pCapture = (CDlgExtra *)m_aExtra[iEx];
			SetCapture( m_hDlg );
			m_pCapture->Invalidate();
			return true;
			}
	return false;
	}

void CDialog :: DragMove( int iX, int iY ){
	if( m_pCapture ) {
		m_pCapture->DragMove( iX, iY );
		m_pCapture->Invalidate();
		}
	}

void CDialog :: DragUp( int iX, int iY ){
	if( m_pCapture ) {
		m_pCapture->DragUp( iX, iY );
		m_pCapture->Invalidate();
		}
	ReleaseCapture();
	m_pCapture = NULL;
	}

CDlgEdit * CDialog :: CreateEdit( int iX, int iY, int iWid, int iHgt, int iID, char * pchTip, DWORD dwStyle ) { 
	HINSTANCE hInst = (HINSTANCE) GetWindowLong( m_hDlg, GWL_HINSTANCE );
	HWND h=CreateWindow( "edit", NULL, WS_CHILD|WS_VISIBLE|WS_BORDER|ES_AUTOHSCROLL|dwStyle, 
			iX, iY, iWid, iHgt, m_hDlg, (HMENU) iID, hInst, NULL );
	CDlgEdit * p = new CDlgEdit;
	if( p != NULL ) {
		m_aControls.AddItem( p );
		p->SetHandle(h);
		
		if( pchTip!=NULL ) 
			p->CreateTip( pchTip );
		}
	
	return p;
	}

CDlgCheck * CDialog :: CreateCheckBox( int iX, int iY, int iWid, int iHgt, int iID, char * pchTip, HFONT fnt, DWORD dwStyle ) { 
	HINSTANCE hInst = (HINSTANCE) GetWindowLong( m_hDlg, GWL_HINSTANCE );
	HWND h=CreateWindow( "button", NULL, WS_CHILD | BS_AUTOCHECKBOX | WS_VISIBLE | dwStyle, 
							iX, iY, iWid, iHgt, m_hDlg, (HMENU) iID, hInst, NULL );
	CDlgCheck * p = new CDlgCheck;
	if( p != NULL ) {
		m_aControls.AddItem( p );
		p->SetHandle(h);
		
		if( pchTip!=NULL ) 
			CreateTip( iX, iY, iX+iWid, iY+iHgt, pchTip );
		if( fnt != NULL ) 
			SendMessage( h, WM_SETFONT, (WPARAM)fnt, (LPARAM)TRUE );
		}
	return p;
	}

CDlgScroll * CDialog :: CreateScrollBar( int iX, int iY, int iWid, int iHgt, int iID, char * pchTip, DWORD dwStyle ) { 
	HINSTANCE hInst = (HINSTANCE) GetWindowLong( m_hDlg, GWL_HINSTANCE );
	HWND h=CreateWindow( "SCROLLBAR", NULL, WS_CHILD | WS_VISIBLE | dwStyle, 
							iX, iY, iWid, iHgt, m_hDlg, (HMENU) iID, hInst, NULL );
	SetScrollPos( h, SB_CTL, 50, TRUE );							
	CDlgScroll * p = new CDlgScroll;
	if( p != NULL ) {
		m_aControls.AddItem( p );
		p->SetHandle(h);
		
		if( pchTip!=NULL ) 
			CreateTip( iX, iY, iX+iWid, iY+iHgt, pchTip );
		}
	return p;
	}

CDlgControl * CDialog :: CreateLabel( int iX, int iY, int iWid, int iHgt, int iID, char * pchTip, HFONT fnt, DWORD dwStyle ) { 
	HINSTANCE hInst = (HINSTANCE) GetWindowLong( m_hDlg, GWL_HINSTANCE );
	HWND h=CreateWindow( "static", NULL, WS_CHILD |WS_VISIBLE | dwStyle/*| SS_WHITEFRAME*/ , 
							iX, iY, iWid, iHgt, m_hDlg, (HMENU) iID, hInst, NULL );
	CDlgControl * p = new CDlgControl;
	if( p != NULL ) {
		m_aControls.AddItem( p );
		p->SetHandle(h);
		
		if( pchTip!=NULL ) 
			CreateTip( iX, iY, iX+iWid, iY+iHgt, pchTip );
		if( fnt != NULL ) 
			SendMessage( h, WM_SETFONT, (WPARAM)fnt, (LPARAM)TRUE );
		}
	return p;
	}

CDlgControl * CDialog :: CreateButton( int iX, int iY, int iWid, int iHgt, int iID, HBITMAP hb, char * pchTip ) { 
	HINSTANCE hInst = (HINSTANCE) GetWindowLong( m_hDlg, GWL_HINSTANCE );
	HWND h=CreateWindow( "button", NULL, WS_CHILD|WS_BORDER|WS_VISIBLE|(hb==NULL?0:BS_BITMAP), 
			iX, iY, iWid, iHgt, m_hDlg, (HMENU) iID, hInst, NULL );
	if( hb!=NULL)
		SendMessage( h, BM_SETIMAGE, IMAGE_BITMAP, (LPARAM) hb );
	CDlgControl * p = new CDlgControl;
	if( p != NULL ) {
		m_aControls.AddItem( p );
		p->SetHandle(h);

		if( pchTip!=NULL ) 
			p->CreateTip( pchTip );
		}
	return p;
	}
	
void CDialog :: CreateTip( int iX0, int iY0, int iX1, int iY1, char * pchMsg ) { 
    // CREATE A TOOLTIP WINDOW 
    if( m_hTip == NULL ) { 
		m_hTip = CreateWindowEx( WS_EX_TOPMOST, TOOLTIPS_CLASS, NULL,
								WS_POPUP | TTS_NOPREFIX | TTS_ALWAYSTIP,		
								CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT,
								m_hDlg, NULL, GetInstance(), NULL );
		SetWindowPos( m_hTip, HWND_TOPMOST, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE | SWP_NOACTIVATE);
		}
	
    // add tip region
    TOOLINFO ti;
    ti.cbSize = sizeof(TOOLINFO);
    ti.uFlags = TTF_SUBCLASS;
    ti.hwnd = m_hDlg;
    ti.hinst = GetInstance();
    ti.uId = 0;

    ti.lpszText = pchMsg;       
    ti.rect.left = iX0;    
    ti.rect.top = iY0;
    ti.rect.right = iX1;
    ti.rect.bottom = iY1;
    SendMessage( m_hTip, TTM_ADDTOOL, 0, (LPARAM) (LPTOOLINFO) &ti );
    }

void CDialog :: ShowControls() { 
	// post show controls to avoid flicker 
	for( int iCtl=0; iCtl<m_aControls.Count(); ++iCtl )
		Control(iCtl)->SetVisible(true);
	}

void CDialog :: DestroyControls() { 
	// post show controls to avoid flicker 
	int iCtls = m_aControls.Count();
	for( int iCtl=0; iCtl<iCtls; ++iCtl )
		DestroyWindow( Control(iCtl)->GetHandle() );
	m_aControls.DeleteAll();
	}

CDlgControl * CDialog :: Control( int iPos ) { 
	return (CDlgControl *)m_aControls[iPos];
	}
	
void CDialog :: DestroyExtra( int iID ) {
	int iExtras=m_aExtra.Count();
	for( int i=0; i<iExtras; ++i ) 
		if( ((CDlgExtra *)m_aExtra[i])->m_iID == iID ) { 
			m_aExtra.DeleteItem( i );
			return;
			}
	}
	
void CDialog :: DestroyControl( int iID ) {
	HWND hCtl = GetDlgItem( m_hDlg, iID );
	if( hCtl == NULL ) 
		return;
	
	int iCtls = m_aControls.Count();
	for( int iCtl=0; iCtl<iCtls; ++iCtl ) 
		if( Control(iCtl)->GetHandle() == hCtl ) {
			DestroyWindow( hCtl );
			m_aControls.DeleteItem(iCtl);
			break;
			}
	}
	
void CDialog :: GetBounds( RECT * prct ) { 
	GetWindowRect( m_hDlg, prct );
	ScreenToClient( m_hDlg, (POINT *)(prct) );
	ScreenToClient( m_hDlg, (POINT *)(&prct->right) );
	}

//------------------------------------------------------------------------------------
/*
//-- return count of windows having specified ID
int WindowListCountID( int iID ) { 
	int iWin=0;
	for( int i=0; i<g_iWinOpen; ++i ) 
		if( ((CDialog *) GetWindowLong( GetParent(g_aWindows[i]), 0 ))->m_iID == iID ) 
			++iWin;
	return iWin;
	}
*/

bool WindowListCanClose( ) { 
	for( int iWin=0; iWin<g_iWinOpen; ++iWin ) {
		CDialog * pDlg = ((CDialog *) GetWindowLong( GetParent(g_aWindows[iWin]), 0 ));
		if( /* pDlg->m_bDocumentDialog && */ ! pDlg->CanClose() ) 
			return false;
		}
	return true;
	}
			
bool WindowListClose( ) { 
	bool bClosed=true;
	for( int iWin=0; iWin<g_iWinOpen; ++iWin ) {
		CDialog * pDlg = ((CDialog *) GetWindowLong( GetParent(g_aWindows[iWin]), 0 ));
		if( ! pDlg->Close() ) 
			bClosed=false;
		}
	return bClosed;
	}

int WindowListFindID( int iID, int iStartPos ) {
	for( int i=max(0,iStartPos); i<g_iWinOpen; ++i ) 
		if( ((CDialog *) GetWindowLong( GetParent(g_aWindows[i]), 0 ))->m_iID == iID ) 
			return i;
	return -1;
	}

bool WindowListAdd( HWND hFrame, HWND hNew ) { 
	//-- Initial allocation of array of HWND's
	if( g_aWindows == NULL ) { 
		g_aWindows = new HWND[g_iWinAlloc=24];
		if( g_aWindows == NULL )
			return false;		// todo - post error message
		}
	//-- Expand the array 
	if( g_iWinOpen == g_iWinAlloc ) { 
		HWND * aH = new HWND[g_iWinAlloc+=24];
		if( aH == NULL ) 
			return false;		// todo - post error message
		for( int iH=0; iH<g_iWinOpen; ++iH ) 
			aH[iH]=g_aWindows[iH];
		delete [] g_aWindows;
		g_aWindows=aH;
		}
		
	//-- Add to the array 
	g_aWindows[g_iWinOpen++]=hNew;

	//-- resize parent to fit if passed 
	if( hFrame != NULL ) { 	
		WINDOWPLACEMENT wp;
		wp.length=sizeof(WINDOWPLACEMENT);
		GetWindowPlacement( hFrame, &wp );

		RECT rct;
		GetClientRect( hNew, &rct );
		AdjustWindowRect( &rct, WS_CAPTION, false ); 					
		MoveWindow( hFrame, wp.rcNormalPosition.left, wp.rcNormalPosition.top, rct.right-rct.left, rct.bottom-rct.top, TRUE );
		}
	
	return true;
	} 

bool WindowListRemove( HWND h ) { 
	for( int iH=0; iH<g_iWinOpen; ++iH )
		if( g_aWindows[iH]==h ) 
			return WindowListRemove( iH );
	return false;
	}
	
bool WindowListRemove( int iH ) { 
	if( iH<0 || iH>=g_iWinOpen ) 
		return false;
	for( int i=iH+1; i<g_iWinOpen; ++i )
		g_aWindows[i-1]=g_aWindows[i];
	--g_iWinOpen;
	return true;
	}

//-- Message Pump check for all open dialogues 
bool WindowListProcessMessage( MSG * pmsg ){ 
	for( int i=0; i<g_iWinOpen; ++i ) 
		if( IsDialogMessage( g_aWindows[i], pmsg) )
			return true;
	return false;
	}
	
HWND WindowListHandle( int iPos ) { 
	if( iPos<0 || iPos>=g_iWinOpen ) 
		return NULL;
	return g_aWindows[iPos];
	}
	
//-----	Child Window - base window in MDI which has a dialog superimposed on it
LRESULT CALLBACK ChildWndProc( HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam ){
	// WM_CREATE 
	// WM_DESTROY
	// WM_MOVE

	if( message==WM_GETMINMAXINFO 
			|| message==WM_SIZE
			|| message==WM_ENTERSIZEMOVE
			|| message==WM_EXITSIZEMOVE 
			|| message==WM_MDIACTIVATE 
			|| message==WM_KEYUP
			|| message==WM_KEYDOWN
			|| message==WM_CHAR		) { 
		LONG l = GetWindowLong(  hWnd, 0 );	// CVw* ptr 
		if( l ) 
			switch( message ) { 
				case WM_CHAR :
					((CDialog*)l)->OnChar(wParam,lParam);
					break;
					
				case WM_KEYDOWN :
					((CDialog*)l)->OnKeyDown(wParam,lParam);
					break;

				case WM_KEYUP :
					((CDialog*)l)->OnKeyUp(wParam,lParam);
					break;

				case WM_GETMINMAXINFO :
					((CDialog*)l)->OnMinMaxInfo((MINMAXINFO*)lParam);
					break;
			
				case WM_SIZE : 
					((CDialog*)l)->OnSize( wParam, &lParam );
					break;
					
				case WM_ENTERSIZEMOVE : 
					//((CDialog*)l)->OnEnterSizing( &lParam );
					((CDialog*)l)->m_bSizing=true;
					break;

				case WM_EXITSIZEMOVE : 
					((CDialog*)l)->OnExitSizing( &lParam );
					((CDialog*)l)->m_bSizing=false;
					InvalidateRect( hWnd, NULL, false ); // needed ? 
					break;
					
				case WM_MDIACTIVATE : 
					((CDialog*)l)->OnMDIActivate( (HWND)wParam, (HWND)lParam );
					break;					
				}
		}

	return DefMDIChildProc( hWnd, message, wParam, lParam);
	}


//------------------------------------------------------------------------------------
CDocumentDialog :: CDocumentDialog( const char * pchFileFilter, const char * pchFileExtension ) /*: CDialog() */ { 
	m_pchFileFilter=pchFileFilter;
	m_pchFileExtension=pchFileExtension;
	m_bUpdated=false;
	m_bDocumentDialog=true;
	}
	
void CDocumentDialog :: SetUpdated() { 
	m_bUpdated=true;
	}

bool CDocumentDialog :: GetSaveAsFileName( ) { 
	char bufName[_MAX_PATH];
	strcpy( bufName, m_cFileName.ConstBuffer() );
	if( ! GetSaveFileName( bufName, m_pchFileFilter, m_pchFileExtension, GetParent(m_hDlg) ) )
		return false;
		
	m_cFileName.Set( bufName );
	m_bUpdated=false;
	return true;
	}	
	
	
	